// This code implements the MD5 message-digest algorithm.
// The algorithm is due to Ron Rivest.  This code was
// written by Colin Plumb in 1993, no copyright is claimed.
// This code is in the public domain; do with it what you wish.
//
// Equivalent code is available from RSA Data Security, Inc.
// This code has been tested against that, and is equivalent,
// except that you don't need to include two pages of legalese
// with every copy.
//
// To compute the message digest of a chunk of bytes, declare an
// MD5Context structure, pass it to MD5Init, call MD5Update as
// needed on buffers full of bytes, and then call MD5Final, which
// will fill a supplied 16-byte array with the digest.
//
// Changed so as no longer to depend on Colin Plumb's `usual.h' header
// definitions; now uses stuff from dpkg's config.h.
//  - Ian Jackson <ian@chiark.greenend.org.uk>.
// Still in the public domain.
//
// MD5.pas: Translated from C to Delphi by Jordan Russell on 2004-03-16.
// Still in the public domain. The original C code was taken from dpkg.
// Slighlty modified by Ketmar // Vampire Avalon. no global code/algo changes,
// though. Still in the public domain.
//
// $jrsoftware: issrc/Projects/MD5.pas,v 1.2 2004/03/16 17:58:14 jr Exp $
// (based on revision 1.15 from local 'md5' repository)
//
// slightly modified by Ketmar // Vampire Avalon:
//   * added a copule of "packed" keywords
//   * no more than 2GB of data can be processed
// Still in the public domain.
//
unit md5U;

interface

type
  TMD5Word = LongWord;
  TMD5Buf = array [0..3] of TMD5Word;
  TMD5In = array [0..15] of TMD5Word;
  TMD5Context = record
    buf: TMD5Buf;
    count: Integer;
    in_: TMD5In;
  end;

  TMD5Digest = packed array [0..15] of Byte;


function MD5Buf (const buffer; len: Integer): TMD5Digest;
function MD5Equal (const dig0, dig1: TMD5Digest): Boolean;
function MD5ToStr (const md5: TMD5Digest): string;
function MD5FromStr (const s: string; out md5: TMD5Digest): Boolean;

procedure MD5Init (var ctx: TMD5Context);
procedure MD5Update (var ctx: TMD5Context; const buffer; len: Integer);
function MD5Final (var ctx: TMD5Context): TMD5Digest;


implementation


const
  hexDigits: array [0..15] of Char = '0123456789abcdef';


function MD5Equal (const dig0, dig1: TMD5Digest): Boolean;
var
  f: Integer;
begin
  result := false;
  for f := Low(TMD5Digest) to High(TMD5Digest) do if dig0[f] <> dig1[f] then exit;
  result := true;
end;


function MD5ToStr (const md5: TMD5Digest): string;
var
  buf: packed array [0..31] of Char;
  p: PChar;
  f: Integer;
begin
  p := @buf;
  for f := 0 to 15 do
  begin
    p^ := hexDigits[md5[f] shr 4]; Inc(p);
    p^ := hexDigits[md5[f] and 15]; Inc(p);
  end;
  SetString(result, buf, 32);
end;

function MD5FromStr (const s: string; out md5: TMD5Digest): Boolean;
var
  c: Char;
  b: Byte;
  f: Integer;
begin
  result := false;
  if Length(s) <> 32 then exit;
  for f := 0 to 15 do
  begin
    c := UpCase(s[f*2+1]); if not (c in ['0'..'9','A'..'F']) then exit;
    b := ord(c)-ord('0'); if b > 9 then Dec(b, 7);
    md5[f] := (b shl 4);
    c := UpCase(s[f*2+2]); if not (c in ['0'..'9','A'..'F']) then exit;
    b := ord(c)-ord('0'); if b > 9 then Dec(b, 7);
    md5[f] := md5[f] or b;
  end;
  result := true;
end;


procedure MD5Transform (var buf: TMD5Buf; const in_: TMD5In);
// The core of the MD5 algorithm, this alters an existing MD5 hash to
// reflect the addition of 16 longwords of new data.  MD5Update blocks
// the data and converts bytes into longwords for this routine.
//
// JR: The macros have been unrolled...
var
  a, b, c, d: TMD5Word;
begin
  a := buf[0]; b := buf[1]; c := buf[2]; d := buf[3];

  // JR: The stuff below was generated using GenTransformCode.dpr
  Inc(a, in_[0]+$D76AA478+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[1]+$E8C7B756+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[2]+$242070DB+(b xor (d and (a xor b))));  c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[3]+$C1BDCEEE+(a xor (c and (d xor a))));  b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[4]+$F57C0FAF+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[5]+$4787C62A+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[6]+$A8304613+(b xor (d and (a xor b))));  c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[7]+$FD469501+(a xor (c and (d xor a))));  b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[8]+$698098D8+(d xor (b and (c xor d))));  a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[9]+$8B44F7AF+(c xor (a and (b xor c))));  d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[10]+$FFFF5BB1+(b xor (d and (a xor b)))); c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[11]+$895CD7BE+(a xor (c and (d xor a)))); b := ((b shl 22) or (b shr (32-22)))+c;
  Inc(a, in_[12]+$6B901122+(d xor (b and (c xor d)))); a := ((a shl 7) or (a shr (32-7)))+b;
  Inc(d, in_[13]+$FD987193+(c xor (a and (b xor c)))); d := ((d shl 12) or (d shr (32-12)))+a;
  Inc(c, in_[14]+$A679438E+(b xor (d and (a xor b)))); c := ((c shl 17) or (c shr (32-17)))+d;
  Inc(b, in_[15]+$49B40821+(a xor (c and (d xor a)))); b := ((b shl 22) or (b shr (32-22)))+c;

  Inc(a, in_[1]+$F61E2562+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[6]+$C040B340+(b xor (c and (a xor b))));  d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[11]+$265E5A51+(a xor (b and (d xor a)))); c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[0]+$E9B6C7AA+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[5]+$D62F105D+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[10]+$02441453+(b xor (c and (a xor b)))); d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[15]+$D8A1E681+(a xor (b and (d xor a)))); c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[4]+$E7D3FBC8+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[9]+$21E1CDE6+(c xor (d and (b xor c))));  a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[14]+$C33707D6+(b xor (c and (a xor b)))); d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[3]+$F4D50D87+(a xor (b and (d xor a))));  c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[8]+$455A14ED+(d xor (a and (c xor d))));  b := ((b shl 20) or (b shr (32-20)))+c;
  Inc(a, in_[13]+$A9E3E905+(c xor (d and (b xor c)))); a := ((a shl 5) or (a shr (32-5)))+b;
  Inc(d, in_[2]+$FCEFA3F8+(b xor (c and (a xor b))));  d := ((d shl 9) or (d shr (32-9)))+a;
  Inc(c, in_[7]+$676F02D9+(a xor (b and (d xor a))));  c := ((c shl 14) or (c shr (32-14)))+d;
  Inc(b, in_[12]+$8D2A4C8A+(d xor (a and (c xor d)))); b := ((b shl 20) or (b shr (32-20)))+c;

  Inc(a, in_[5]+$FFFA3942+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[8]+$8771F681+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[11]+$6D9D6122+(d xor a xor b)); c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[14]+$FDE5380C+(c xor d xor a)); b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[1]+$A4BEEA44+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[4]+$4BDECFA9+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[7]+$F6BB4B60+(d xor a xor b));  c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[10]+$BEBFBC70+(c xor d xor a)); b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[13]+$289B7EC6+(b xor c xor d)); a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[0]+$EAA127FA+(a xor b xor c));  d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[3]+$D4EF3085+(d xor a xor b));  c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[6]+$04881D05+(c xor d xor a));  b := ((b shl 23) or (b shr (32-23)))+c;
  Inc(a, in_[9]+$D9D4D039+(b xor c xor d));  a := ((a shl 4) or (a shr (32-4)))+b;
  Inc(d, in_[12]+$E6DB99E5+(a xor b xor c)); d := ((d shl 11) or (d shr (32-11)))+a;
  Inc(c, in_[15]+$1FA27CF8+(d xor a xor b)); c := ((c shl 16) or (c shr (32-16)))+d;
  Inc(b, in_[2]+$C4AC5665+(c xor d xor a));  b := ((b shl 23) or (b shr (32-23)))+c;

  Inc(a, in_[0]+$F4292244+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[7]+$432AFF97+(b xor (a or (not c))));  d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[14]+$AB9423A7+(a xor (d or (not b)))); c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[5]+$FC93A039+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[12]+$655B59C3+(c xor (b or (not d)))); a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[3]+$8F0CCC92+(b xor (a or (not c))));  d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[10]+$FFEFF47D+(a xor (d or (not b)))); c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[1]+$85845DD1+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[8]+$6FA87E4F+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[15]+$FE2CE6E0+(b xor (a or (not c)))); d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[6]+$A3014314+(a xor (d or (not b))));  c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[13]+$4E0811A1+(d xor (c or (not a)))); b := ((b shl 21) or (b shr (32-21)))+c;
  Inc(a, in_[4]+$F7537E82+(c xor (b or (not d))));  a := ((a shl 6) or (a shr (32-6)))+b;
  Inc(d, in_[11]+$BD3AF235+(b xor (a or (not c)))); d := ((d shl 10) or (d shr (32-10)))+a;
  Inc(c, in_[2]+$2AD7D2BB+(a xor (d or (not b))));  c := ((c shl 15) or (c shr (32-15)))+d;
  Inc(b, in_[9]+$EB86D391+(d xor (c or (not a))));  b := ((b shl 21) or (b shr (32-21)))+c;

  Inc(buf[0], a); Inc(buf[1], b); Inc(buf[2], c); Inc(buf[3], d);
end;


procedure MD5Init (var ctx: TMD5Context);
// Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
// initialization constants.
begin
  ctx.buf[0] := TMD5Word($67452301);
  ctx.buf[1] := TMD5Word($EFCDAB89);
  ctx.buf[2] := TMD5Word($98BADCFE);
  ctx.buf[3] := TMD5Word($10325476);
  ctx.count := 0;
end;

procedure MD5Update (var ctx: TMD5Context; const buffer; len: Integer);
// Update context to reflect the concatenation of another buffer full
// of bytes.
var
  buf: ^Byte;
  t: Integer;
begin
  if len < 1 then exit;

  buf := @buffer;

  // Update byte count
  t := ctx.count; Inc(ctx.count, len);

  t := 64-(t and $3F); // Space available in ctx.in (at least 1)
  if t > len then
  begin
    Move(buf^, Pointer(Integer(@ctx.in_)+64-t)^, len);
    exit;
  end;

  // First chunk is an odd size
  Move(buf^, Pointer(Integer(@ctx.in_)+64-t)^, t);
  MD5Transform(ctx.buf, ctx.in_);
  Inc(buf, t); Dec(len, t);

  // Process data in 64-byte chunks
  while len >= 64 do
  begin
    Move(buf^, ctx.in_, 64);
    MD5Transform(ctx.buf, ctx.in_);
    Inc(buf, 64); Dec(len, 64);
  end;

  // Handle any remaining bytes of data
  if len > 0 then Move(buf^, ctx.in_, len);
end;

function MD5Final (var ctx: TMD5Context): TMD5Digest;
// Final wrapup - pad to 64-byte boundary with the bit pattern
// 1 0* (64-bit count of bits processed, MSB-first)
var
  count: Integer;
  p: ^Byte;
begin
  count := ctx.count and $3F; // Number of bytes in ctx.in
  p := @ctx.in_; Inc(p, count);

  // Set the first char of padding to 0x80. There is always room.
  p^ := $80;
  Inc(p);

  // Bytes of padding needed to make 56 bytes (-8..55)
  count := 56-1-count;

  if count < 0 then
  begin
    // Padding forces an extra block
    FillChar(p^, count+8, 0);
    MD5Transform(ctx.buf, ctx.in_);
    p := @ctx.in_;
    count := 56;
  end;
  FillChar(p^, count, 0);

  // Append length in bits and transform
  ctx.in_[14] := LongWord(ctx.count) shl 3;
  ctx.in_[15] := LongWord(ctx.count) shr 29;
  MD5Transform(ctx.buf, ctx.in_);

  Move(ctx.buf, result, 16);
  FillChar(ctx, SizeOf(ctx), 0);  // In case it's sensitive
end;

// New functions by JR:
function MD5Buf (const buffer; len: Integer): TMD5Digest;
var
  context: TMD5Context;
begin
  MD5Init(context);
  if len > 0 then MD5Update(context, buffer, len);
  result := MD5Final(context);
end;


end.
